# -*- coding: utf-8 -*-

import logging
import os
from logging.handlers import SysLogHandler

from prometheus_client import start_http_server
from telegram_handler import TelegramHandler

from scdlbot.scdlbot import ScdlBot

logging_handlers = []

console_formatter = logging.Formatter("[%(name)s] %(levelname)s: %(message)s")
console_handler = logging.StreamHandler()
console_handler.setFormatter(console_formatter)
console_handler.setLevel(logging.DEBUG)
logging_handlers.append(console_handler)

tg_bot_token = os.environ["TG_BOT_TOKEN"]
alert_chat_ids = list(map(int, os.getenv("ALERT_CHAT_IDS", "0").split(",")))
telegram_handler = TelegramHandler(token=tg_bot_token, chat_id=str(alert_chat_ids[0]))
telegram_handler.setLevel(logging.WARNING)
logging_handlers.append(telegram_handler)

syslog_debug = bool(int(os.getenv("SYSLOG_DEBUG", "0")))
syslog_logging_level = logging.DEBUG if syslog_debug else logging.INFO
syslog_hostname = os.getenv("HOSTNAME", "test-host")
syslog_formatter = logging.Formatter("%(asctime)s " + syslog_hostname + " %(name)s: %(message)s", datefmt="%b %d %H:%M:%S")

syslog_address = os.getenv("SYSLOG_ADDRESS", "")
if syslog_address:
    syslog_host, syslog_udp_port = syslog_address.split(":")
    syslog_handler = SysLogHandler(address=(syslog_host, int(syslog_udp_port)))
    syslog_handler.setFormatter(syslog_formatter)
    syslog_handler.setLevel(syslog_logging_level)
    logging_handlers.append(syslog_handler)

logging.basicConfig(format="%(asctime)s - %(name)s - %(levelname)s - %(message)s", datefmt="%Y-%m-%d %H:%M:%S", level=logging.DEBUG, handlers=logging_handlers)


def main():
    # expose prometheus/openmetrics metrics:
    metrics_host = os.getenv("METRICS_HOST", "127.0.0.1")
    metrics_port = int(os.getenv("METRICS_PORT", "8000"))
    start_http_server(metrics_port, addr=metrics_host)

    store_chat_id = int(os.getenv("STORE_CHAT_ID", "0"))
    no_flood_chat_ids = list(map(int, os.getenv("NO_FLOOD_CHAT_IDS", "0").split(",")))
    dl_timeout = int(os.getenv("DL_TIMEOUT", "300"))
    dl_dir = os.path.expanduser(os.getenv("DL_DIR", "/tmp/scdlbot"))
    chat_storage_file = os.path.expanduser(os.getenv("CHAT_STORAGE", "/tmp/scdlbotdata"))
    serve_audio = bool(int(os.getenv("SERVE_AUDIO", "0")))
    app_url = os.getenv("APP_URL", "")
    tg_bot_api = os.getenv("TG_BOT_API", "https://api.telegram.org")
    max_tg_file_size = int(os.getenv("MAX_TG_FILE_SIZE", "45_000_000"))
    max_convert_file_size = int(os.getenv("MAX_CONVERT_FILE_SIZE", "80_000_000"))
    proxies = os.getenv("PROXIES", None)
    if proxies:
        proxies = proxies.split(",")
    source_ips = os.getenv("SOURCE_IPS", None)
    if source_ips:
        source_ips = source_ips.split(",")
    cookies_file = os.getenv("COOKIES_FILE", "")
    workers = int(os.getenv("WORKERS", 4))

    scdlbot = ScdlBot(
        tg_bot_token,
        tg_bot_api,
        proxies,
        store_chat_id,
        no_flood_chat_ids,
        alert_chat_ids,
        dl_dir,
        dl_timeout,
        max_tg_file_size,
        max_convert_file_size,
        chat_storage_file,
        app_url,
        serve_audio,
        cookies_file,
        source_ips,
        workers,
    )

    use_webhook = bool(int(os.getenv("USE_WEBHOOK", "0")))
    webhook_host = os.getenv("HOST", "127.0.0.1")
    webhook_port = int(os.getenv("PORT", "5000"))
    cert_file = os.getenv("CERT_FILE", "")
    cert_key_file = os.getenv("CERT_KEY_FILE", "")
    url_path = os.getenv("URL_PATH", tg_bot_token.replace(":", ""))
    scdlbot.start(use_webhook, webhook_host, webhook_port, cert_file, cert_key_file, url_path)


if __name__ == "__main__":
    main()



# Configuration file for the Sphinx documentation builder.
#
# This file does only contain a selection of the most common options. For a
# full list see the documentation:
# http://www.sphinx-doc.org/en/master/config

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.

import os
import sys

import tomlkit

sys.path.insert(0, os.path.abspath(".."))


# -- Project information -----------------------------------------------------


def _get_project_meta():
    with open("../pyproject.toml") as pyproject:
        file_contents = pyproject.read()

    return tomlkit.parse(file_contents)["tool"]["poetry"]


pkg_meta = _get_project_meta()
project = str(pkg_meta["name"])
copyright = "2020, gpchelkin"  # noqa: A001  # skipcq: PYL-W0622
author = "gpchelkin"

# The short X.Y version
version = str(pkg_meta["version"])
# The full version, including alpha/beta/rc tags
release = str(version)

# -- General configuration ---------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
needs_sphinx = "3.2"

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.doctest",
    "sphinx.ext.todo",
    "sphinx.ext.coverage",
    "sphinx.ext.viewcode",
    "sphinx.ext.autosummary",
    # Used to write beautiful docstrings:
    "sphinx.ext.napoleon",
    # Used to include .md files:
    "m2r2",
    # Used to insert typehints into the final docs:
    "sphinx_autodoc_typehints",
]

autoclass_content = "class"
autodoc_member_order = "bysource"

autodoc_member_order = "bysource"
autodoc_default_flags = {
    "members": "",
    "undoc-members": "code,error_template",
    "exclude-members": "__dict__,__weakref__",
}

# Set `typing.TYPE_CHECKING` to `True`:
# https://pypi.org/project/sphinx-autodoc-typehints/
set_type_checking_flag = False

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:

source_suffix = [".rst", ".md"]

# The master toctree document.
master_doc = "index"

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = "en"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path .
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

add_module_names = False

autodoc_default_options = {
    "show-inheritance": True,
}

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = "alabaster"

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
html_theme_options = {
    "sidebar_collapse": False,
    "show_powered_by": False,
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

# Custom sidebar templates, must be a dictionary that maps document names
# to template names.
#
# This is required for the alabaster theme
# refs: http://alabaster.readthedocs.io/en/latest/installation.html#sidebars
html_sidebars = {
    "**": [
        "about.html",
        "badges.html",
        "navigation.html",
        "moreinfo.html",
        "github.html",
        "searchbox.html",
    ],
}

# -- Extension configuration -------------------------------------------------

napoleon_numpy_docstring = False

# -- Options for todo extension ----------------------------------------------

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = True



# -*- coding: utf-8 -*-

"""Main module."""

import gc
import pathlib
import random
import shelve
import shutil
from datetime import datetime
from multiprocessing import Process, Queue
from queue import Empty
from subprocess import PIPE, TimeoutExpired  # skipcq: BAN-B404
from urllib.parse import urljoin, urlparse
from uuid import uuid4

import ffmpeg
from boltons.urlutils import find_all_links
from mutagen.id3 import ID3
from mutagen.mp3 import EasyMP3 as MP3
from prometheus_client import Summary
from telegram import Chat, ChatAction, ChatMember, InlineKeyboardButton, InlineKeyboardMarkup, Message, MessageEntity, Update
from telegram.error import BadRequest, ChatMigrated, NetworkError, TelegramError, TimedOut, Unauthorized
from telegram.ext import CallbackContext, CallbackQueryHandler, CommandHandler, Filters, MessageHandler, Updater
from telegram.ext.dispatcher import run_async

from scdlbot.utils import *

logger = logging.getLogger(__name__)

REQUEST_TIME = Summary("request_processing_seconds", "Time spent processing request")


class ScdlBot:
    def __init__(
        self,
        tg_bot_token,
        tg_bot_api="https://api.telegram.org",
        proxies=None,
        store_chat_id=None,
        no_flood_chat_ids=None,
        alert_chat_ids=None,
        dl_dir="/tmp/scdlbot",
        dl_timeout=300,
        max_tg_file_size=45_000_000,
        max_convert_file_size=80_000_000,
        chat_storage_file="/tmp/scdlbotdata",
        app_url=None,
        serve_audio=False,
        cookies_file=None,
        source_ips=None,
        workers=4,
    ):
        self.SITES = {
            "sc": "soundcloud",
            "scapi": "api.soundcloud",
            "bc": "bandcamp",
            "yt": "youtu",
        }
        self.APP_URL = app_url
        self.DL_TIMEOUT = dl_timeout
        self.TG_BOT_API = tg_bot_api
        self.MAX_TG_FILE_SIZE = max_tg_file_size
        self.MAX_CONVERT_FILE_SIZE = max_convert_file_size
        self.SERVE_AUDIO = serve_audio
        if self.SERVE_AUDIO:
            self.MAX_TG_FILE_SIZE = 19_000_000
        self.HELP_TEXT = get_response_text("help.tg.md")
        self.SETTINGS_TEXT = get_response_text("settings.tg.md")
        self.DL_TIMEOUT_TEXT = get_response_text("dl_timeout.txt").format(self.DL_TIMEOUT // 60)
        self.WAIT_BIT_TEXT = [get_response_text("wait_bit.txt"), get_response_text("wait_beat.txt"), get_response_text("wait_beet.txt")]
        self.NO_AUDIO_TEXT = get_response_text("no_audio.txt")
        self.NO_URLS_TEXT = get_response_text("no_urls.txt")
        self.OLD_MSG_TEXT = get_response_text("old_msg.txt")
        self.REGION_RESTRICTION_TEXT = get_response_text("region_restriction.txt")
        self.DIRECT_RESTRICTION_TEXT = get_response_text("direct_restriction.txt")
        self.LIVE_RESTRICTION_TEXT = get_response_text("live_restriction.txt")
        # self.chat_storage = {}
        self.chat_storage = shelve.open(chat_storage_file, writeback=True)
        for chat_id in no_flood_chat_ids:
            self.init_chat(chat_id=chat_id, chat_type=Chat.PRIVATE if chat_id > 0 else Chat.SUPERGROUP, flood="no")
        self.ALERT_CHAT_IDS = set(alert_chat_ids) if alert_chat_ids else set()
        self.STORE_CHAT_ID = store_chat_id
        self.DL_DIR = dl_dir
        self.COOKIES_DOWNLOAD_FILE = "/tmp/scdlbot_cookies.txt"
        self.proxies = proxies
        self.source_ips = source_ips
        # https://yandex.com/support/music-app-ios/search-and-listen/listening-abroad.html
        self.cookies_file = cookies_file
        self.workers = workers

        # if sc_auth_token:
        #     config = configparser.ConfigParser()
        #     config['scdl'] = {}
        #     config['scdl']['path'] = self.DL_DIR
        #     config['scdl']['auth_token'] = sc_auth_token
        #     config_dir = os.path.join(os.path.expanduser('~'), '.config', 'scdl')
        #     config_path = os.path.join(config_dir, 'scdl.cfg')
        #     os.makedirs(config_dir, exist_ok=True)
        #     with open(config_path, 'w') as config_file:
        #         config.write(config_file)

        self.updater = Updater(token=tg_bot_token, base_url=f"{self.TG_BOT_API}/bot", use_context=True, base_file_url=f"{self.TG_BOT_API}/file/bot", workers=self.workers)
        dispatcher = self.updater.dispatcher

        start_command_handler = CommandHandler("start", self.help_command_callback)
        dispatcher.add_handler(start_command_handler)
        help_command_handler = CommandHandler("help", self.help_command_callback)
        dispatcher.add_handler(help_command_handler)
        settings_command_handler = CommandHandler("settings", self.settings_command_callback)
        dispatcher.add_handler(settings_command_handler)

        dl_command_handler = CommandHandler("dl", self.common_command_callback, filters=~Filters.update.edited_message & ~Filters.forwarded)
        dispatcher.add_handler(dl_command_handler)
        link_command_handler = CommandHandler("link", self.common_command_callback, filters=~Filters.update.edited_message & ~Filters.forwarded)
        dispatcher.add_handler(link_command_handler)
        message_with_links_handler = MessageHandler(
            ~Filters.update.edited_message
            & ~Filters.command
            & (
                (Filters.text & (Filters.entity(MessageEntity.URL) | Filters.entity(MessageEntity.TEXT_LINK)))
                | (Filters.caption & (Filters.caption_entity(MessageEntity.URL) | Filters.caption_entity(MessageEntity.TEXT_LINK)))
            ),
            self.common_command_callback,
        )
        dispatcher.add_handler(message_with_links_handler)

        button_query_handler = CallbackQueryHandler(self.button_query_callback)
        dispatcher.add_handler(button_query_handler)

        unknown_handler = MessageHandler(Filters.command, self.unknown_command_callback)
        dispatcher.add_handler(unknown_handler)

        dispatcher.add_error_handler(self.error_callback)

        self.bot_username = self.updater.bot.get_me().username
        self.RANT_TEXT_PRIVATE = "Read /help to learn how to use me"
        self.RANT_TEXT_PUBLIC = "[Start me in PM to read help and learn how to use me](t.me/{}?start=1)".format(self.bot_username)

    def start(self, use_webhook=False, webhook_host="127.0.0.1", webhook_port=None, cert_file=None, cert_key_file=None, url_path="scdlbot"):
        if use_webhook:
            self.updater.start_webhook(listen=webhook_host, port=webhook_port, url_path=url_path)
            # cert=cert_file if cert_file else None,
            # key=cert_key_file if cert_key_file else None,
            # webhook_url=urljoin(app_url, url_path))
            self.updater.bot.set_webhook(url=urljoin(self.APP_URL, url_path), certificate=open(cert_file, "rb") if cert_file else None)
        else:
            self.updater.start_polling()
        logger.warning("Bot started")
        self.updater.idle()

    def unknown_command_callback(self, update: Update, context: CallbackContext):
        pass
        # bot.send_message(chat_id=update.message.chat_id, text="Unknown command")

    def error_callback(self, update: Update, context: CallbackContext):  # skipcq: PYL-R0201
        try:
            raise context.error
        except Unauthorized:
            # remove update.message.chat_id from conversation list
            logger.debug("Update {} caused Unauthorized error: {}".format(update, context.error))
        except BadRequest:
            # handle malformed requests - read more below!
            logger.debug("Update {} caused BadRequest error: {}".format(update, context.error))
        except TimedOut:
            # handle slow connection problems
            logger.debug("Update {} caused TimedOut error: {}".format(update, context.error))
        except NetworkError:
            # handle other connection problems
            logger.debug("Update {} caused NetworkError: {}".format(update, context.error))
        except ChatMigrated as e:
            # the chat_id of a group has changed, use e.new_chat_id instead
            logger.debug("Update {} caused ChatMigrated error: {}".format(update, context.error))
        except TelegramError:
            # handle all other telegram related errors
            logger.debug("Update {} caused TelegramError: {}".format(update, context.error))

    def init_chat(self, message=None, chat_id=None, chat_type=None, flood="yes"):
        if message:
            chat_id = str(message.chat_id)
            chat_type = message.chat.type
        else:
            chat_id = str(chat_id)
        if chat_id not in self.chat_storage:
            self.chat_storage[chat_id] = {}
        if "settings" not in self.chat_storage[chat_id]:
            self.chat_storage[chat_id]["settings"] = {}
        if "mode" not in self.chat_storage[chat_id]["settings"]:
            if chat_type == Chat.PRIVATE:
                self.chat_storage[chat_id]["settings"]["mode"] = "dl"
            else:
                self.chat_storage[chat_id]["settings"]["mode"] = "ask"
        if "flood" not in self.chat_storage[chat_id]["settings"]:
            self.chat_storage[chat_id]["settings"]["flood"] = flood
        if "rant_msg_ids" not in self.chat_storage[chat_id]["settings"]:
            self.chat_storage[chat_id]["settings"]["rant_msg_ids"] = []
        self.chat_storage.sync()
        # logger.debug("Current chat_storage: %r", self.chat_storage)

    def cleanup_chat(self, chat_id):
        chat_msgs = self.chat_storage[str(chat_id)].copy()
        for msg_id in chat_msgs:
            if msg_id != "settings":
                timedelta = datetime.now().replace(tzinfo=None) - self.chat_storage[str(chat_id)][msg_id]["message"].date.replace(tzinfo=None)
                if timedelta.days > 0:
                    self.chat_storage[str(chat_id)].pop(msg_id)
        self.chat_storage.sync()

    def rant_and_cleanup(self, bot, chat_id, rant_text, reply_to_message_id=None):
        rant_msg = bot.send_message(chat_id=chat_id, reply_to_message_id=reply_to_message_id, text=rant_text, parse_mode="Markdown", disable_web_page_preview=True)
        flood = self.chat_storage[str(chat_id)]["settings"]["flood"]
        if flood == "no":
            rant_msgs = self.chat_storage[str(chat_id)]["settings"]["rant_msg_ids"].copy()
            for rant_msg_id in rant_msgs:
                try:
                    bot.delete_message(chat_id=chat_id, message_id=rant_msg_id)
                except:
                    pass
                self.chat_storage[str(chat_id)]["settings"]["rant_msg_ids"].remove(rant_msg_id)
            self.chat_storage[str(chat_id)]["settings"]["rant_msg_ids"].append(rant_msg.message_id)
            self.chat_storage.sync()

    def help_command_callback(self, update: Update, context: CallbackContext):
        self.init_chat(update.message)
        event_name = "help"
        entities = update.message.parse_entities(types=[MessageEntity.BOT_COMMAND])
        for entity_value in entities.values():
            event_name = entity_value.replace("/", "").replace("@{}".format(self.bot_username), "")
            break
        log_and_track(event_name, update.message)
        chat_id = update.message.chat_id
        chat_type = update.message.chat.type
        reply_to_message_id = update.message.message_id
        flood = self.chat_storage[str(chat_id)]["settings"]["flood"]
        if chat_type != Chat.PRIVATE and flood == "no":
            self.rant_and_cleanup(context.bot, chat_id, self.RANT_TEXT_PUBLIC, reply_to_message_id=reply_to_message_id)
        else:
            context.bot.send_message(chat_id=chat_id, text=self.HELP_TEXT, parse_mode="Markdown", disable_web_page_preview=True)

    def get_wait_text(self):
        return random.choice(self.WAIT_BIT_TEXT)

    def get_settings_inline_keyboard(self, chat_id):
        mode = self.chat_storage[str(chat_id)]["settings"]["mode"]
        flood = self.chat_storage[str(chat_id)]["settings"]["flood"]
        emoji_yes = "✅"
        emoji_no = "❌"
        button_dl = InlineKeyboardButton(text=" ".join([emoji_yes if mode == "dl" else emoji_no, "Download"]), callback_data=" ".join(["settings", "dl"]))
        button_link = InlineKeyboardButton(text=" ".join([emoji_yes if mode == "link" else emoji_no, "Links"]), callback_data=" ".join(["settings", "link"]))
        button_ask = InlineKeyboardButton(text=" ".join([emoji_yes if mode == "ask" else emoji_no, "Ask"]), callback_data=" ".join(["settings", "ask"]))
        button_flood = InlineKeyboardButton(text=" ".join([emoji_yes if flood == "yes" else emoji_no, "Captions"]), callback_data=" ".join(["settings", "flood"]))
        button_close = InlineKeyboardButton(text=" ".join([emoji_no, "Close settings"]), callback_data=" ".join(["settings", "close"]))
        inline_keyboard = InlineKeyboardMarkup([[button_dl, button_link, button_ask], [button_flood, button_close]])
        return inline_keyboard

    def settings_command_callback(self, update: Update, context: CallbackContext):
        self.init_chat(update.message)
        log_and_track("settings")
        chat_id = update.message.chat_id
        context.bot.send_message(chat_id=chat_id, parse_mode="Markdown", reply_markup=self.get_settings_inline_keyboard(chat_id), text=self.SETTINGS_TEXT)

    def common_command_callback(self, update: Update, context: CallbackContext):
        self.init_chat(update.message)
        chat_id = update.message.chat_id
        chat_type = update.message.chat.type
        reply_to_message_id = update.message.message_id
        command_entities = update.message.parse_entities(types=[MessageEntity.BOT_COMMAND])
        command_passed = False
        if not command_entities:
            command_passed = False
            # if no command then it is just a message and use default mode
            mode = self.chat_storage[str(chat_id)]["settings"]["mode"]
        else:
            command_passed = True
            # try to determine mode from command
            mode = None
            for entity_value in command_entities.values():
                mode = entity_value.replace("/", "").replace("@{}".format(self.bot_username), "")
                break
            if not mode:
                mode = "dl"
        if command_passed and not context.args:
            rant_text = self.RANT_TEXT_PRIVATE if chat_type == Chat.PRIVATE else self.RANT_TEXT_PUBLIC
            rant_text += "\nYou can simply send message with links (to download) OR command as `/{} <links>`.".format(mode)
            self.rant_and_cleanup(context.bot, chat_id, rant_text, reply_to_message_id=reply_to_message_id)
            return
        event_name = ("{}_cmd".format(mode)) if command_passed else ("{}_msg".format(mode))
        log_and_track(event_name, update.message)

        apologize = False
        # apologize and send TYPING: always in PM, only when it's command in non-PM
        if chat_type == Chat.PRIVATE or command_passed:
            apologize = True
        source_ip = None
        proxy = None
        if self.source_ips:
            source_ip = random.choice(self.source_ips)
        if self.proxies:
            proxy = random.choice(self.proxies)
        self.prepare_urls(
            message=update.message, mode=mode, source_ip=source_ip, proxy=proxy, apologize=apologize, chat_id=chat_id, reply_to_message_id=reply_to_message_id, bot=context.bot
        )

    def button_query_callback(self, update: Update, context: CallbackContext):
        btn_msg = update.callback_query.message
        self.init_chat(btn_msg)
        user_id = update.callback_query.from_user.id
        btn_msg_id = btn_msg.message_id
        chat = btn_msg.chat
        chat_id = chat.id
        chat_type = chat.type
        orig_msg_id, action = update.callback_query.data.split()
        if orig_msg_id == "settings":
            if chat_type != Chat.PRIVATE:
                chat_member_status = chat.get_member(user_id).status
                if chat_member_status not in [ChatMember.ADMINISTRATOR, ChatMember.CREATOR] and user_id not in self.ALERT_CHAT_IDS:
                    log_and_track("settings_fail")
                    update.callback_query.answer(text="You're not chat admin")
                    return
            log_and_track("settings_{}".format(action), btn_msg)
            if action == "close":
                context.bot.delete_message(chat_id, btn_msg_id)
            else:
                setting_changed = False
                if action in ["dl", "link", "ask"]:
                    current_setting = self.chat_storage[str(chat_id)]["settings"]["mode"]
                    if action != current_setting:
                        setting_changed = True
                        self.chat_storage[str(chat_id)]["settings"]["mode"] = action
                elif action in ["flood"]:
                    current_setting = self.chat_storage[str(chat_id)]["settings"]["flood"]
                    setting_changed = True
                    self.chat_storage[str(chat_id)]["settings"][action] = "no" if current_setting == "yes" else "yes"
                if setting_changed:
                    self.chat_storage.sync()
                    update.callback_query.answer(text="Settings changed")
                    update.callback_query.edit_message_reply_markup(parse_mode="Markdown", reply_markup=self.get_settings_inline_keyboard(chat_id))
                else:
                    update.callback_query.answer(text="Settings not changed")

        elif orig_msg_id in self.chat_storage[str(chat_id)]:
            msg_from_storage = self.chat_storage[str(chat_id)].pop(orig_msg_id)
            orig_msg = msg_from_storage["message"]
            urls = msg_from_storage["urls"]
            source_ip = msg_from_storage["source_ip"]
            proxy = msg_from_storage["proxy"]
            log_and_track("{}_msg".format(action), orig_msg)
            if action == "dl":
                update.callback_query.answer(text=self.get_wait_text())
                wait_message = update.callback_query.edit_message_text(parse_mode="Markdown", text=get_italic(self.get_wait_text()))
                for url in urls:
                    self.download_url_and_send(
                        context.bot, url, urls[url], chat_id=chat_id, reply_to_message_id=orig_msg_id, wait_message_id=wait_message.message_id, source_ip=source_ip, proxy=proxy
                    )
            elif action == "link":
                context.bot.send_message(chat_id=chat_id, reply_to_message_id=orig_msg_id, parse_mode="Markdown", disable_web_page_preview=True, text=get_link_text(urls))
                context.bot.delete_message(chat_id=chat_id, message_id=btn_msg_id)
            elif action == "nodl":
                context.bot.delete_message(chat_id=chat_id, message_id=btn_msg_id)
        else:
            update.callback_query.answer(text=self.OLD_MSG_TEXT)
            context.bot.delete_message(chat_id=chat_id, message_id=btn_msg_id)

    @REQUEST_TIME.time()
    @run_async
    def prepare_urls(self, message, mode=None, source_ip=None, proxy=None, apologize=None, chat_id=None, reply_to_message_id=None, bot=None):
        direct_urls = False
        if mode == "link":
            direct_urls = True

        if apologize:
            bot.send_chat_action(chat_id=chat_id, action=ChatAction.TYPING)

        if isinstance(message, Message):
            urls = []
            url_entities = message.parse_entities(types=[MessageEntity.URL])
            url_caption_entities = message.parse_caption_entities(types=[MessageEntity.URL])
            url_entities.update(url_caption_entities)
            for entity in url_entities:
                url_str = url_entities[entity]
                logger.debug("Entity URL Parsed: %s", url_str)
                if "://" not in url_str:
                    url_str = "http://{}".format(url_str)
                urls.append(URL(url_str))
            text_link_entities = message.parse_entities(types=[MessageEntity.TEXT_LINK])
            text_link_caption_entities = message.parse_caption_entities(types=[MessageEntity.TEXT_LINK])
            text_link_entities.update(text_link_caption_entities)
            for entity in text_link_entities:
                url_str = entity.url
                logger.debug("Entity Text Link Parsed: %s", url_str)
                urls.append(URL(url_str))
        else:
            urls = find_all_links(message, default_scheme="http")
        logger.debug(urls)

        urls_dict = {}
        for url_item in urls:
            # unshorten soundcloud.app.goo.gl and other links, but not tiktok or instagram or youtube:
            if "tiktok" in url_item.host or "instagr" in url_item.host or self.SITES["yt"] in url_item.host:
                url = url_item
            else:
                try:
                    url = URL(
                        requests.head(
                            url_item,
                            allow_redirects=True,
                            timeout=5,
                            proxies=dict(http=proxy, https=proxy),
                            headers={"User-Agent": "Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:105.0) Gecko/20100101 Firefox/105.0"},
                        ).url
                    )
                except:
                    url = url_item
            url_text = url.to_text(True)
            # FIXME crutch:
            url_text = url_text.replace("m.soundcloud.com", "soundcloud.com")
            url_parts_num = len([part for part in url.path_parts if part])
            try:
                if (
                    # SoundCloud: tracks, sets and widget pages, no /you/ pages  # TODO private sets are 5
                    (self.SITES["sc"] in url.host and (2 <= url_parts_num <= 4 or self.SITES["scapi"] in url_text) and (not "you" in url.path_parts))
                    or
                    # Bandcamp: tracks and albums
                    (self.SITES["bc"] in url.host and (2 <= url_parts_num <= 2))
                    or
                    # YouTube: videos and playlists
                    (self.SITES["yt"] in url.host and ("youtu.be" in url.host or "watch" in url.path or "playlist" in url.path))
                ):
                    if direct_urls or self.SITES["yt"] in url.host:
                        urls_dict[url_text] = get_direct_urls(url_text, self.cookies_file, self.COOKIES_DOWNLOAD_FILE, source_ip, proxy)
                    else:
                        urls_dict[url_text] = "http"
                elif not any((site in url.host for site in self.SITES.values())):
                    urls_dict[url_text] = get_direct_urls(url_text, self.cookies_file, self.COOKIES_DOWNLOAD_FILE, source_ip, proxy)
            except ProcessExecutionError:
                logger.debug("youtube-dl get-url failed: %s", url_text)
            except URLError as exc:
                urls_dict[url_text] = exc.status

        logger.debug(urls_dict)
        if not urls_dict and apologize:
            bot.send_message(chat_id=chat_id, reply_to_message_id=reply_to_message_id, text=self.NO_URLS_TEXT, parse_mode="Markdown")
            return

        if mode == "dl":
            wait_message = bot.send_message(chat_id=chat_id, reply_to_message_id=reply_to_message_id, parse_mode="Markdown", text=get_italic(self.get_wait_text()))
            for url in urls_dict:
                self.download_url_and_send(
                    bot, url, urls_dict[url], chat_id=chat_id, reply_to_message_id=reply_to_message_id, wait_message_id=wait_message.message_id, source_ip=source_ip, proxy=proxy
                )
        elif mode == "link":
            wait_message = bot.send_message(chat_id=chat_id, reply_to_message_id=reply_to_message_id, parse_mode="Markdown", text=get_italic(self.get_wait_text()))
            bot.send_message(chat_id=chat_id, reply_to_message_id=reply_to_message_id, parse_mode="Markdown", disable_web_page_preview=True, text=get_link_text(urls_dict))
            bot.delete_message(chat_id=chat_id, message_id=wait_message.message_id)
        elif mode == "ask":
            # ask only if good urls exist
            if "http" in " ".join(urls_dict.values()):
                orig_msg_id = str(reply_to_message_id)
                self.chat_storage[str(chat_id)][orig_msg_id] = {"message": message, "urls": urls_dict, "source_ip": source_ip, "proxy": proxy}
                question = "🎶 links found, what to do?"
                button_dl = InlineKeyboardButton(text="✅ Download", callback_data=" ".join([orig_msg_id, "dl"]))
                button_link = InlineKeyboardButton(text="❇️ Links", callback_data=" ".join([orig_msg_id, "link"]))
                button_cancel = InlineKeyboardButton(text="❎", callback_data=" ".join([orig_msg_id, "nodl"]))
                inline_keyboard = InlineKeyboardMarkup([[button_dl, button_link, button_cancel]])
                bot.send_message(chat_id=chat_id, reply_to_message_id=reply_to_message_id, reply_markup=inline_keyboard, text=question)
            self.cleanup_chat(chat_id)

    @REQUEST_TIME.time()
    @run_async
    def download_url_and_send(self, bot, url, direct_urls, chat_id, reply_to_message_id=None, wait_message_id=None, source_ip=None, proxy=None):
        bot.send_chat_action(chat_id=chat_id, action=ChatAction.RECORD_AUDIO)
        download_dir = os.path.join(self.DL_DIR, str(uuid4()))
        shutil.rmtree(download_dir, ignore_errors=True)
        os.makedirs(download_dir)

        status = 0
        if direct_urls == "direct":
            status = -3
        elif direct_urls == "country":
            status = -4
        elif direct_urls == "live":
            status = -5
        elif direct_urls == "timeout":
            status = -6
        else:
            if (self.SITES["sc"] in url and self.SITES["scapi"] not in url) or (self.SITES["bc"] in url):
                cmd_name = "scdl"
                cmd_args = []
                cmd = None
                cmd_input = None
                if self.SITES["sc"] in url and self.SITES["scapi"] not in url:
                    cmd = scdl_bin
                    cmd_name = str(cmd)
                    cmd_args = (
                        "-l",
                        url,  # URL of track/playlist/user
                        "-c",  # Continue if a music already exist
                        "--path",
                        download_dir,  # Download the music to a custom path
                        "--onlymp3",  # Download only the mp3 file even if the track is Downloadable
                        "--addtofile",  # Add the artist name to the filename if it isn't in the filename already
                        "--addtimestamp",
                        # Adds the timestamp of the creation of the track to the title (useful to sort chronologically)
                        "--no-playlist-folder",
                        # Download playlist tracks into directory, instead of making a playlist subfolder
                        "--extract-artist",  # Set artist tag from title instead of username
                    )
                    cmd_input = None
                elif self.SITES["bc"] in url:
                    cmd = bandcamp_dl_bin
                    cmd_name = str(cmd)
                    cmd_args = (
                        "--base-dir",
                        download_dir,  # Base location of which all files are downloaded
                        "--template",
                        "%{track} - %{artist} - %{title} [%{album}]",  # Output filename template
                        "--overwrite",  # Overwrite tracks that already exist
                        "--group",  # Use album/track Label as iTunes grouping
                        # "--embed-art",  # Embed album art (if available)
                        "--no-slugify",  # Disable slugification of track, album, and artist names
                        url,  # URL of album/track
                    )
                    cmd_input = "yes"

                logger.info("%s starts: %s", cmd_name, url)
                env = None
                if proxy:
                    env = {"http_proxy": proxy, "https_proxy": proxy}
                cmd_proc = cmd[cmd_args].popen(env=env, stdin=PIPE, stdout=PIPE, stderr=PIPE, universal_newlines=True)
                try:
                    cmd_stdout, cmd_stderr = cmd_proc.communicate(input=cmd_input, timeout=self.DL_TIMEOUT)
                    cmd_retcode = cmd_proc.returncode
                    # TODO listed are common scdl problems for one track with 0 retcode, all its output is always in stderr:
                    if cmd_retcode or (any(err in cmd_stderr for err in ["Error resolving url", "is not streamable", "Failed to get item"]) and ".mp3" not in cmd_stderr):
                        raise ProcessExecutionError(cmd_args, cmd_retcode, cmd_stdout, cmd_stderr)
                    logger.info("%s succeeded: %s", cmd_name, url)
                    status = 1
                except TimeoutExpired:
                    cmd_proc.kill()
                    logger.info("%s took too much time and dropped: %s", cmd_name, url)
                    status = -1
                except ProcessExecutionError:
                    logger.exception("%s failed: %s", cmd_name, url)

        if status == 0:
            cmd = youtube_dl_func
            cmd_name = "youtube_dl_func"
            # TODO: set different ydl_opts for different sites
            host = urlparse(url).hostname
            ydl_opts = {}
            if host == "tiktok.com" or host.endswith(".tiktok.com"):
                ydl_opts = {
                    "outtmpl": os.path.join(download_dir, "tiktok.%(ext)s"),
                    "videoformat": "mp4",
                }
            elif "instagr" in host:
                ydl_opts = {
                    "outtmpl": os.path.join(download_dir, "inst.%(ext)s"),
                    "videoformat": "mp4",
                    "postprocessors": [
                        {
                            "key": "FFmpegVideoConvertor",
                            "preferedformat": "mp4",
                        }
                    ],
                }
            else:
                ydl_opts = {
                    "outtmpl": os.path.join(download_dir, "%(title)s.%(ext)s"),
                    # default: %(autonumber)s - %(title)s-%(id)s.%(ext)s
                    "format": "bestaudio/best",
                    "postprocessors": [
                        {
                            "key": "FFmpegExtractAudio",
                            "preferredcodec": "mp3",
                            "preferredquality": "128",
                        },
                        # {'key': 'EmbedThumbnail',}, {'key': 'FFmpegMetadata',},
                    ],
                    "noplaylist": True,
                }
            if proxy:
                ydl_opts["proxy"] = proxy
            if source_ip:
                ydl_opts["source_address"] = source_ip
            # https://github.com/ytdl-org/youtube-dl/blob/master/youtube_dl/YoutubeDL.py#L210
            if self.cookies_file:
                if "http" in self.cookies_file:
                    ydl_opts["cookiefile"] = self.COOKIES_DOWNLOAD_FILE
                else:
                    ydl_opts["cookiefile"] = self.cookies_file
            queue = Queue()
            cmd_args = (
                url,
                ydl_opts,
                queue,
            )
            logger.info("%s starts: %s", cmd_name, url)
            cmd_proc = Process(target=cmd, args=cmd_args)
            cmd_proc.start()
            try:
                cmd_retcode, cmd_stderr = queue.get(block=True, timeout=self.DL_TIMEOUT)
                cmd_stdout = ""
                cmd_proc.join()
                if cmd_retcode:
                    raise ProcessExecutionError(cmd_args, cmd_retcode, cmd_stdout, cmd_stderr)
                    # raise cmd_status  # TODO: pass and re-raise original Exception?
                logger.info("%s succeeded: %s", cmd_name, url)
                status = 1
            except Empty:
                cmd_proc.join(1)
                if cmd_proc.is_alive():
                    cmd_proc.terminate()
                logger.info("%s took too much time and dropped: %s", cmd_name, url)
                status = -1
            except ProcessExecutionError:
                logger.exception("%s failed: %s", cmd_name, url)
                status = -2
            gc.collect()

        if status in [-1, -6]:
            bot.send_message(chat_id=chat_id, reply_to_message_id=reply_to_message_id, text=self.DL_TIMEOUT_TEXT, parse_mode="Markdown")
        elif status == -2:
            bot.send_message(chat_id=chat_id, reply_to_message_id=reply_to_message_id, text=self.NO_AUDIO_TEXT, parse_mode="Markdown")
        elif status == -3:
            bot.send_message(chat_id=chat_id, reply_to_message_id=reply_to_message_id, text=self.DIRECT_RESTRICTION_TEXT, parse_mode="Markdown")
        elif status == -4:
            bot.send_message(chat_id=chat_id, reply_to_message_id=reply_to_message_id, text=self.REGION_RESTRICTION_TEXT, parse_mode="Markdown")
        elif status == -5:
            bot.send_message(chat_id=chat_id, reply_to_message_id=reply_to_message_id, text=self.LIVE_RESTRICTION_TEXT, parse_mode="Markdown")
        elif status == 1:
            file_list = []
            for d, dirs, files in os.walk(download_dir):
                for file in files:
                    file_list.append(os.path.join(d, file))
            if not file_list:
                logger.info("No files in dir: %s", download_dir)
                bot.send_message(chat_id=chat_id, reply_to_message_id=reply_to_message_id, text="*Sorry*, I couldn't download any files from provided links", parse_mode="Markdown")
            else:
                for file in sorted(file_list):
                    file_name = os.path.split(file)[-1]
                    file_parts = []
                    try:
                        file_root, file_ext = os.path.splitext(file)
                        file_format = file_ext.replace(".", "").lower()
                        file_size = os.path.getsize(file)
                        if file_format not in ["mp3", "m4a", "mp4"]:
                            raise FileNotSupportedError(file_format)
                        if file_size > self.MAX_CONVERT_FILE_SIZE:
                            raise FileTooLargeError(file_size)
                        # FIXME tiktok.mp4 is for tiktok, inst.mp4 for instagram
                        if file_format not in ["mp3"] and not ("tiktok." in file or "inst." in file):
                            logger.info("Converting: %s", file)
                            try:
                                file_converted = file.replace(file_ext, ".mp3")
                                ffinput = ffmpeg.input(file)
                                ffmpeg.output(ffinput, file_converted, audio_bitrate="128k", vn=None).run()
                                file = file_converted
                                file_root, file_ext = os.path.splitext(file)
                                file_format = file_ext.replace(".", "").lower()
                                file_size = os.path.getsize(file)
                            except Exception:
                                # TODO exceptions
                                raise FileNotConvertedError

                        file_parts = []
                        if file_size <= self.MAX_TG_FILE_SIZE:
                            file_parts.append(file)
                        else:
                            logger.info("Splitting: %s", file)
                            id3 = None
                            try:
                                id3 = ID3(file, translate=False)
                            except:
                                pass

                            parts_number = file_size // self.MAX_TG_FILE_SIZE + 1

                            # https://github.com/c0decracker/video-splitter
                            # https://superuser.com/a/1354956/464797
                            try:
                                # file_duration = float(ffmpeg.probe(file)['format']['duration'])
                                part_size = file_size // parts_number
                                cur_position = 0
                                for i in range(parts_number):
                                    file_part = file.replace(file_ext, ".part{}{}".format(str(i + 1), file_ext))
                                    ffinput = ffmpeg.input(file)
                                    if i == (parts_number - 1):
                                        ffmpeg.output(ffinput, file_part, codec="copy", vn=None, ss=cur_position).run()
                                    else:
                                        ffmpeg.output(ffinput, file_part, codec="copy", vn=None, ss=cur_position, fs=part_size).run()
                                        part_duration = float(ffmpeg.probe(file_part)["format"]["duration"])
                                        cur_position += part_duration
                                    if id3:
                                        try:
                                            id3.save(file_part, v1=2, v2_version=4)
                                        except:
                                            pass
                                    file_parts.append(file_part)
                            except Exception:
                                # TODO exceptions
                                raise FileSplittedPartiallyError(file_parts)

                    except FileNotSupportedError as exc:
                        if not (exc.file_format in ["m3u", "jpg", "jpeg", "png", "finished", "tmp"]):
                            logger.warning("Unsupported file format: %s", file_name)
                            bot.send_message(
                                chat_id=chat_id,
                                reply_to_message_id=reply_to_message_id,
                                text="*Sorry*, downloaded file `{}` is in format I could not yet convert or send".format(file_name),
                                parse_mode="Markdown",
                            )
                    except FileTooLargeError as exc:
                        logger.info("Large file for convert: %s", file_name)
                        bot.send_message(
                            chat_id=chat_id,
                            reply_to_message_id=reply_to_message_id,
                            text="*Sorry*, downloaded file `{}` is `{}` MB and it is larger than I could convert (`{} MB`)".format(
                                file_name, exc.file_size // 1000000, self.MAX_CONVERT_FILE_SIZE // 1000000
                            ),
                            parse_mode="Markdown",
                        )
                    except FileSplittedPartiallyError as exc:
                        file_parts = exc.file_parts
                        logger.exception("Splitting failed: %s", file_name)
                        bot.send_message(
                            chat_id=chat_id,
                            reply_to_message_id=reply_to_message_id,
                            text="*Sorry*, not enough memory to convert file `{}`..".format(file_name),
                            parse_mode="Markdown",
                        )
                    except FileNotConvertedError as exc:
                        logger.exception("Splitting failed: %s", file_name)
                        bot.send_message(
                            chat_id=chat_id,
                            reply_to_message_id=reply_to_message_id,
                            text="*Sorry*, not enough memory to convert file `{}`..".format(file_name),
                            parse_mode="Markdown",
                        )
                    try:
                        caption = None
                        flood = self.chat_storage[str(chat_id)]["settings"]["flood"]
                        if flood == "yes":
                            addition = ""
                            url_obj = URL(url)
                            if self.SITES["yt"] in url_obj.host:
                                source = "YouTube"
                                file_root, file_ext = os.path.splitext(file_name)
                                file_title = file_root.replace(file_ext, "")
                                addition = ": " + file_title
                            elif self.SITES["sc"] in url_obj.host:
                                source = "SoundCloud"
                            elif self.SITES["bc"] in url_obj.host:
                                source = "Bandcamp"
                            else:
                                source = url_obj.host.replace(".com", "").replace("www.", "").replace("m.", "")
                            # if "youtu.be" in url_obj.host:
                            #     url = url.replace("http://", "").replace("https://", "")
                            # else:
                            #     url = shorten_url(url)
                            caption = "@{} _got it from_ [{}]({}){}".format(self.bot_username.replace("_", "\_"), source, url, addition.replace("_", "\_"))
                            # logger.info(caption)
                        reply_to_message_id_send = reply_to_message_id if flood == "yes" else None
                        sent_audio_ids = []
                        for index, file_part in enumerate(file_parts):
                            path = pathlib.Path(file_part)
                            file_name = os.path.split(file_part)[-1]
                            # file_name = translit(file_name, 'ru', reversed=True)
                            logger.info("Sending: %s", file_name)
                            bot.send_chat_action(chat_id=chat_id, action=ChatAction.UPLOAD_AUDIO)
                            caption_part = None
                            if len(file_parts) > 1:
                                caption_part = "Part {} of {}".format(str(index + 1), str(len(file_parts)))
                            if caption:
                                if caption_part:
                                    caption_full = caption_part + " | " + caption
                                else:
                                    caption_full = caption
                            else:
                                if caption_part:
                                    caption_full = caption_part
                                else:
                                    caption_full = ""
                            # caption_full = textwrap.shorten(caption_full, width=190, placeholder="..")
                            for i in range(3):
                                try:
                                    if file_part.endswith(".mp3"):
                                        mp3 = MP3(file_part)
                                        duration = round(mp3.info.length)
                                        performer = None
                                        title = None
                                        try:
                                            performer = ", ".join(mp3["artist"])
                                            title = ", ".join(mp3["title"])
                                        except:
                                            pass
                                        if "127.0.0.1" in self.TG_BOT_API:
                                            audio = path.absolute().as_uri()
                                            logger.debug(audio)
                                        elif self.SERVE_AUDIO:
                                            audio = str(urljoin(self.APP_URL, str(path.relative_to(self.DL_DIR))))
                                            logger.debug(audio)
                                        else:
                                            audio = open(file_part, "rb")
                                        if i > 0:
                                            # maybe: Reply message not found
                                            reply_to_message_id_send = None
                                        audio_msg = bot.send_audio(
                                            chat_id=chat_id,
                                            reply_to_message_id=reply_to_message_id_send,
                                            audio=audio,
                                            duration=duration,
                                            performer=performer,
                                            title=title,
                                            caption=caption_full,
                                            parse_mode="Markdown",
                                        )
                                        sent_audio_ids.append(audio_msg.audio.file_id)
                                        logger.info("Sending succeeded: %s", file_name)
                                        break
                                    elif "tiktok." in file_part or "inst." in file_part:
                                        video = open(file_part, "rb")
                                        duration = float(ffmpeg.probe(file_part)["format"]["duration"])
                                        videostream = next(item for item in ffmpeg.probe(file_part)["streams"] if item["codec_type"] == "video")
                                        width = int(videostream["width"])
                                        height = int(videostream["height"])
                                        video_msg = bot.send_video(
                                            chat_id=chat_id,
                                            reply_to_message_id=reply_to_message_id_send,
                                            video=video,
                                            supports_streaming=True,
                                            duration=duration,
                                            width=width,
                                            height=height,
                                            caption=caption_full,
                                            parse_mode="Markdown",
                                        )
                                        sent_audio_ids.append(video_msg.video.file_id)
                                        logger.info("Sending succeeded: %s", file_name)
                                        break
                                except TelegramError:
                                    if i == 2:
                                        logger.exception("Sending failed because of TelegramError: %s", file_name)
                        if len(sent_audio_ids) != len(file_parts):
                            raise FileSentPartiallyError(sent_audio_ids)

                    except FileSentPartiallyError as exc:
                        sent_audio_ids = exc.sent_audio_ids
                        bot.send_message(
                            chat_id=chat_id,
                            reply_to_message_id=reply_to_message_id,
                            text="*Sorry*, could not send file `{}` or some of it's parts..".format(file_name),
                            parse_mode="Markdown",
                        )
                        logger.warning("Sending some parts failed: %s", file_name)

        if not self.SERVE_AUDIO:
            shutil.rmtree(download_dir, ignore_errors=True)
        if wait_message_id:  # TODO: delete only once
            try:
                bot.delete_message(chat_id=chat_id, message_id=wait_message_id)
            except:
                pass



import logging
import os

import pkg_resources
import requests

try:
    import yt_dlp as youtube_dl

    youtube_dl_bin_name = "yt-dlp"
except:
    try:
        import youtube_dl

        youtube_dl_bin_name = "youtube-dl"
    except:
        import youtube_dlc as youtube_dl

        youtube_dl_bin_name = "youtube-dlc"

from boltons.urlutils import URL
from plumbum import ProcessExecutionError, ProcessTimedOut, local

from scdlbot.exceptions import *

# from requests.exceptions import Timeout, RequestException, SSLError

bin_path = os.getenv("BIN_PATH", "")
scdl_bin = local[os.path.join(bin_path, "scdl")]
bandcamp_dl_bin = local[os.path.join(bin_path, "bandcamp-dl")]
youtube_dl_bin = local[os.path.join(bin_path, youtube_dl_bin_name)]

BOTAN_TRACK_URL = "https://api.botan.io/track"

logger = logging.getLogger(__name__)


def get_response_text(file_name):
    # https://stackoverflow.com/a/20885799/2490759
    path = "/".join(("texts", file_name))
    return pkg_resources.resource_string(__name__, path).decode("UTF-8")


def get_direct_urls(url, cookies_file=None, cookies_download_file=None, source_ip=None, proxy=None):
    logger.debug("Entered get_direct_urls")
    youtube_dl_args = []

    # https://github.com/ytdl-org/youtube-dl#how-do-i-pass-cookies-to-youtube-dl
    if cookies_file:
        if "http" in cookies_file:
            try:
                r = requests.get(cookies_file, allow_redirects=True, timeout=5)
                open(cookies_download_file, "wb").write(r.content)
                youtube_dl_args.extend(["--cookies", cookies_download_file])
            except:
                pass
        else:
            youtube_dl_args.extend(["--cookies", cookies_file])

    if source_ip:
        youtube_dl_args.extend(["--source-address", source_ip])

    if proxy:
        youtube_dl_args.extend(["--proxy", proxy])

    youtube_dl_args.extend(["--get-url", url])
    try:
        ret_code, std_out, std_err = youtube_dl_bin[youtube_dl_args].run(timeout=60)
    except ProcessTimedOut as exc:
        raise URLTimeoutError
    except ProcessExecutionError as exc:
        # TODO: look at case: one page has multiple videos, some available, some not
        if "returning it as such" in exc.stderr:
            raise URLDirectError
        if "proxy server" in exc.stderr:
            raise URLCountryError
        raise exc
    if "yt_live_broadcast" in std_out:
        raise URLLiveError
    return std_out


def get_italic(text):
    return "_{}_".format(text)


def youtube_dl_func(url, ydl_opts, queue=None):
    ydl = youtube_dl.YoutubeDL(ydl_opts)
    try:
        ydl.download([url])
    except Exception as exc:
        ydl_status = 1, str(exc)
        # ydl_status = exc  #TODO: pass and re-raise original Exception
    else:
        ydl_status = 0, "OK"
    if queue:
        queue.put(ydl_status)
    else:
        return ydl_status


# def botan_track(token, message, event_name):
#     try:
#         # uid = message.chat_id
#         uid = message.from_user.id
#     except AttributeError:
#         logger.warning('Botan no chat_id in message')
#         return False
#     num_retries = 2
#     ssl_verify = True
#     for i in range(num_retries):
#         try:
#             r = requests.post(
#                 BOTAN_TRACK_URL,
#                 params={"token": token, "uid": uid, "name": event_name},
#                 data=message.to_json(),
#                 verify=ssl_verify,
#                 timeout=2,
#             )
#             return r.json()
#         except Timeout:
#             logger.exception("Botan timeout on event: %s", event_name)
#         except SSLError:
#             ssl_verify = False
#         except (Exception, RequestException, ValueError):
#             # catastrophic error
#             logger.exception("Botan 🙀astrophic error on event: %s", event_name)
#     return False


def shorten_url(url):
    try:
        return requests.get("https://clck.ru/--?url=" + url).text.replace("https://", "")
    except:
        return url


def log_and_track(event_name, message=None):
    logger.info("Event: %s", event_name)
    if message:
        pass
        # if self.botan_token:
        #     return botan_track(self.botan_token, message, event_name)


def get_link_text(urls):
    link_text = ""
    for i, url in enumerate(urls):
        link_text += "[Source Link #{}]({}) | `{}`\n".format(str(i + 1), url, URL(url).host)
        direct_urls = urls[url].splitlines()
        for direct_url in direct_urls:
            if "http" in direct_url:
                content_type = ""
                if "googlevideo" in direct_url:
                    if "audio" in direct_url:
                        content_type = "Audio"
                    else:
                        content_type = "Video"
                # direct_url = shorten_url(direct_url)
                link_text += "• {} [Direct Link]({})\n".format(content_type, direct_url)
    return link_text



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['scdlbot']

package_data = \
{'': ['*'], 'scdlbot': ['texts/*']}

install_requires = \
['bandcamp-downloader==0.0.13',
 'boltons==21.0.0',
 'ffmpeg-python==0.2.0',
 'mutagen>=1.46.0,<2.0.0',
 'patool==1.12',
 'plumbum>=1.8.0,<2.0.0',
 'prometheus_client>=0.15.0,<0.16.0',
 'python-telegram-bot==12.8',
 'python-telegram-handler==2.2.1',
 'requests[socks]>=2.28.1,<3.0.0',
 'scdl==2.7.3',
 'yt-dlp==2022.11.11']

entry_points = \
{'console_scripts': ['scdlbot = scdlbot.__main__:main']}

setup_kwargs = {
    'name': 'scdlbot',
    'version': '0.14.1',
    'description': 'Telegram Bot for downloading MP3 rips of tracks/sets from SoundCloud, Bandcamp, YouTube with tags and artwork',
    'long_description': 'Music Downloader Telegram Bot aka scdlbot\n=========================================\n\n\n.. image:: https://img.shields.io/badge/telegram-@scdlbot-blue.svg\n        :target: https://t.me/scdlbot\n        :alt: Telegram Bot\n\n.. image:: https://img.shields.io/github/license/gpchelkin/scdlbot.svg\n        :target: https://github.com/gpchelkin/scdlbot/blob/master/LICENSE\n        :alt: MIT License\n\n.. image:: https://readthedocs.org/projects/scdlbot/badge/?version=latest\n        :target: https://scdlbot.readthedocs.io/\n        :alt: Documentation Status\n\n.. image:: https://img.shields.io/pypi/v/scdlbot.svg\n        :target: https://pypi.org/project/scdlbot\n        :alt: PyPI Version\n\n.. image:: https://github.com/gpchelkin/scdlbot/workflows/build/badge.svg?branch=master&event=push\n        :target: https://github.com/gpchelkin/scdlbot/actions/workflows/build.yml\n        :alt: GitHub Actions Build Status\n\n.. image:: https://github.com/gpchelkin/scdlbot/actions/workflows/codeql-analysis.yml/badge.svg?branch=master&event=push\n        :target: https://github.com/gpchelkin/scdlbot/actions/workflows/codeql-analysis.yml\n        :alt: GitHub Actions CodeQL Status\n\n.. image:: https://deepsource.io/gh/gpchelkin/scdlbot.svg/?label=active+issues&show_trend=true\n        :target: https://deepsource.io/gh/gpchelkin/scdlbot/?ref=repository-badge\n        :alt: DeepSource Active Issues\n\n.. image:: https://deepsource.io/gh/gpchelkin/scdlbot.svg/?label=resolved+issues&show_trend=true\n        :target: https://deepsource.io/gh/gpchelkin/scdlbot/?ref=repository-badge\n        :alt: DeepSource Resolved Issues\n\n.. image:: https://codeclimate.com/github/gpchelkin/scdlbot/badges/issue_count.svg\n        :target: https://codeclimate.com/github/gpchelkin/scdlbot\n        :alt: Code Climate Issue Count\n\n.. image:: https://api.codacy.com/project/badge/Grade/7dfb6d8e7a094987b303e9283fc7368c\n        :target: https://app.codacy.com/gh/gpchelkin/scdlbot\n        :alt: Codacy Build Status\n\n.. image:: https://codebeat.co/badges/57243b9d-2269-4f31-a35b-6aedd11626d2\n        :target: https://codebeat.co/projects/github-com-gpchelkin-scdlbot-master\n        :alt: Codebeat Quality\n\n.. image:: https://www.codefactor.io/repository/github/gpchelkin/scdlbot/badge\n        :target: https://www.codefactor.io/repository/github/gpchelkin/scdlbot\n        :alt: CodeFactor\n\nTelegram Bot for downloading MP3 rips of tracks/sets from\nSoundCloud, Bandcamp, YouTube with tags and artwork.\n\n\n* Free software: `MIT License <https://github.com/gpchelkin/scdlbot/blob/master/LICENSE>`__\n* Documentation: https://scdlbot.readthedocs.io\n\n\n.. contents:: :depth: 2\n\n\nscdlbot Usage in Telegram\n-------------------------\n\nSend ``/start`` or ``/help`` command to bot\nor refer directly to the `help message <scdlbot/texts/help.tg.md>`__.\n\nPlease report all bugs and issues and suggest your improvements\nto `issues <https://github.com/gpchelkin/scdlbot/issues>`__.\n\nSupported sites and mainly used packages\n^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^\n\nscdlbot is standing on the shoulders of giants:\n\n-  `Telegram Bot API <https://core.telegram.org/bots/api>`__:\n   `python-telegram-bot <https://github.com/python-telegram-bot/python-telegram-bot>`__\n-  `SoundCloud <https://soundcloud.com>`__:\n   `scdl <https://github.com/flyingrub/scdl>`__\n-  `Bandcamp <https://bandcamp.com>`__:\n   `bandcamp-dl <https://github.com/iheanyi/bandcamp-dl>`__\n-  `YouTube <https://www.youtube.com/>`__, `Yandex.Music <https://music.yandex.com/>`__,\n   `Mixcloud <https://www.mixcloud.com/>`__, and almost everything from this `list <https://yt-dl.org/supportedsites.html>`__:\n   `youtube-dl <https://yt-dl.org>`__\n\nRun your own scdlbot\n--------------------\n\nInstallation & Configuration\n^^^^^^^^^^^^^^^^^^^^^^^^^^^^\n\nRequirements\n""""""""""""\n\nThose should be available in your ``PATH``:\n\n-  `Python 3.9+ <https://www.python.org>`__\n   (`pyenv <https://github.com/pyenv/pyenv>`__ and `poetry <https://python-poetry.org/>`__ are recommended)\n-  `FFmpeg 4.0+ <https://ffmpeg.org/download.html>`__ if not running on Heroku\n   (fresh builds for\n   `Windows <https://www.gyan.dev/ffmpeg/builds/#release-builds>`__,\n   `macOS <https://evermeet.cx/ffmpeg/>`__\n   and `Linux <https://johnvansickle.com/ffmpeg/>`__ are recommended)\n-  `Heroku CLI <https://cli.heroku.com>`__ is recommended if you want to deploy to Heroku\n\nInstall / Update stable from `PyPI <https://pypi.org/project/scdlbot>`__ (recommended)\n""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""\n\n::\n\n    pip3 install scdlbot --upgrade\n\n...or get unstable from `Git source repository <https://github.com/gpchelkin/scdlbot>`__\n""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""\n\n::\n\n    git clone https://github.com/gpchelkin/scdlbot.git\n    cd scdlbot\n    pip3 install --requirement requirements.txt\n\n    # Update:\n    git pull\n    pip3 install --requirement requirements.txt\n\n\nConfigure Bot\n"""""""""""""\n\nDownload or copy configuration file sample and set up\nconfig environment variables in it:\n\n::\n\n    # If you\'ve installed from PyPI - download sample config somewhere:\n    curl -o .env https://raw.githubusercontent.com/gpchelkin/scdlbot/master/.env.sample\n\n    # If you\'ve got Git source - just copy sample config:\n    cp .env.sample .env\n\n    # Use your favourite editor. Sample config is self-documented:\n    nano .env\n\nTelegram Bot Settings\n"""""""""""""""""""""\n\nSend the commands from respective filenames in ``telegram_settings`` dir to `@BotFather <https://t.me/BotFather>`__, choose your bot and copy corresponding values in order to use the bot conveniently.\nDisable privacy mode if you want bot to read and check every message in group for links.\nOtherwise, it would work only for commands.\n\nRunning Locally or at Dedicated Server\n^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^\n\nUsing `Heroku Local <https://devcenter.heroku.com/articles/heroku-local#run-your-app-locally-using-the-heroku-local-command-line-tool>`__ (preferred)\n"""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""\n\nYou will need `Heroku CLI <https://cli.heroku.com/>`__ installed.\n\n::\n\n    # If you\'ve installed from PyPI - download Procfile first (otherwise already present in Git repository):\n    curl -O https://raw.githubusercontent.com/gpchelkin/scdlbot/master/Procfile\n\n    # For long polling mode (when USE_WEBHOOK=0):\n    heroku local -e .env worker\n    # For webhook mode (when USE_WEBHOOK=1):\n    heroku local -e .env web\n\nUsing Python only\n"""""""""""""""""\n\n::\n\n    export $(grep -v \'^#\' .env | xargs)\n    python3 -m scdlbot\n    # or in one line:\n    env $(grep -v \'^#\' .env | xargs) python3 -m scdlbot\n\n    # If you\'ve installed package from PyPI into the system,\n    # you can also replace \'python3 -m scdlbot\' with only \'scdlbot\'\n\nDeploying to `Heroku <https://heroku.com/>`__\n^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^\n\n|Deploy|\n\nRegister on Heroku, press the button above and\nconfigure variables for deploying.\nWhen app is deployed you **must** set only one dyno working on\n"Resources" tab in your app settings depending on `which way of getting\nupdates <https://core.telegram.org/bots/api#getting-updates>`__ you have\nchosen and set in config variables: ``worker`` for polling or ``web``\nfor webhook.\n\nManually\n""""""""\n\nYou can do the same as the button above but using `Heroku\nCLI <https://cli.heroku.com/>`__, not as much of a fun. Assuming you are in\n``scdlbot`` repository directory:\n\n::\n\n    heroku login\n    # Create app with Python 3 buildpack and set it for upcoming builds:\n    heroku create --buildpack heroku/python\n    heroku buildpacks:set heroku/python\n    # Add FFmpeg buildpack needed for youtube-dl & scdl:\n    heroku buildpacks:add --index 1 https://github.com/jonathanong/heroku-buildpack-ffmpeg-latest.git --app scdlbot\n    # Deploy app to Heroku:\n    git push heroku master\n    # Set config vars automatically from your local .env file\n    heroku plugins:install heroku-config\n    heroku config:push\n    # Or set them manually:\n    heroku config:set TG_BOT_TOKEN="<TG_BOT_TOKEN>" STORE_CHAT_ID="<STORE_CHAT_ID>" ...\n\nIf you use webhook, start web dyno and stop worker dyno:\n\n::\n\n    heroku ps:scale web=1 worker=0\n    heroku ps:stop worker\n\nIf you use polling, start worker dyno and stop web dyno:\n\n::\n\n    heroku ps:scale worker=1 web=0\n    heroku ps:stop web\n\nSome useful commands:\n\n::\n\n    # Attach to logs:\n    heroku logs -t\n    # Test run ffprobe\n    heroku run "ffprobe -version"\n\nDeploying to `Dokku <https://github.com/dokku/dokku>`__\n^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^\n\nUse Dokku (your own Heroku) installed on your own server.\nApp is tested and fully ready for deployment with polling\n(no webhook yet).\nhttps://github.com/dokku/dokku-letsencrypt\n\n::\n\n    export DOKKU=<your_dokku_server>\n    scp .env $DOKKU:~\n    ssh $DOKKU\n        export DOKKU=<your_dokku_server>\n        dokku apps:create scdlbot\n        dokku certs:generate scdlbot scdlbot.$DOKKU\n        dokku config:set scdlbot $(cat .env | xargs)\n        logout\n    git remote add dokku dokku@$DOKKU:scdlbot\n    git push dokku master\n    ssh $DOKKU\n        dokku ps:scale scdlbot worker=1 web=0\n        dokku ps:restart scdlbot\n\n.. |Deploy| image:: https://www.herokucdn.com/deploy/button.svg\n    :target: https://heroku.com/deploy\n',
    'author': 'George Pchelkin',
    'author_email': 'george@pchelk.in',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/gpchelkin/scdlbot',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)
